#include "eglbackend.h"

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <EGL/eglext.h>

static const char *eglErrorToString(EGLint e)
{
#define CASE(name) case name: return #name
	switch (e) {
		CASE(EGL_SUCCESS);
		CASE(EGL_NOT_INITIALIZED);
		CASE(EGL_BAD_ACCESS);
		CASE(EGL_BAD_ALLOC);
		CASE(EGL_BAD_ATTRIBUTE);
		CASE(EGL_BAD_CONTEXT);
		CASE(EGL_BAD_CONFIG);
		CASE(EGL_BAD_CURRENT_SURFACE);
		CASE(EGL_BAD_DISPLAY);
		default: return "<unknown>";
	}
#undef CASE
}

static void exitEglError(const char *what)
{
	EGLint e = eglGetError();
	fprintf(stderr, "EGL error %d (%s): %s\n", e, eglErrorToString(e), what);
	exit(1);
}

/* attributes for a double buffered visual in RGBA format with at least
* 4 bits per color */
static const EGLint config_attribs[] = {
	EGL_RED_SIZE,             4,
	EGL_GREEN_SIZE,           4,
	EGL_BLUE_SIZE,            4,
	EGL_RENDERABLE_TYPE,      EGL_OPENGL_BIT,
	EGL_NONE,
};

VisualID EGLBackend::initialize(Display *xDisplay)
{
	if (display == EGL_NO_DISPLAY) { // this implies that the context is also unitialized
		// get connection and bind API
		EGLint eglMajor, eglMinor;
		display = eglGetDisplay(xDisplay);
		if (display == EGL_NO_DISPLAY)                         
			exitEglError("Failed to get EGL display");
		if (eglInitialize(display, &eglMajor, &eglMinor) == EGL_FALSE)                   
			exitEglError("Failed to initialize EGL");
		printf("Using EGL version %d.%d\n", eglMajor, eglMinor);
		eglBindAPI(EGL_OPENGL_API);
		// get an appropriate config
		EGLConfig configs[1];
		EGLint count;
		if (eglChooseConfig(display, config_attribs, configs, 1, &count) == EGL_FALSE || count == 0)                            
			exitEglError("Failed to choose config");
		config = configs[0];
	}
	// return visual ID
	EGLint val;
	eglGetConfigAttrib(display, config, EGL_NATIVE_VISUAL_ID, &val);
	return (VisualID)val;
}

void EGLBackend::createContext(Window window)
{
	assert(display != EGL_NO_DISPLAY && context == EGL_NO_CONTEXT);
	surface = eglCreateWindowSurface(display, config, window, NULL);
	/* create an EGL context and use it with the surface */                                    
	context = eglCreateContext(display, config, EGL_NO_CONTEXT, NULL);
	if (context == EGL_NO_CONTEXT)
		exitEglError("Failed to create context");
	if (eglMakeCurrent(display, surface, surface, context) == EGL_FALSE)
		exitEglError("Failed to make context current");
}

EGLBackend::~EGLBackend()
{
	if (display != EGL_NO_DISPLAY) {
		if (context != EGL_NO_CONTEXT) {
			eglMakeCurrent(display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
			eglDestroyContext(display, context);
			eglDestroySurface(display, surface);
		}
		eglTerminate(display);
		eglReleaseThread();
	}
}

void EGLBackend::swapBuffers() const
{
	assert(context != EGL_NO_CONTEXT); // this implies the display is also initialized
	eglSwapBuffers(display, surface);
}

void EGLBackend::setSwapInterval(int i) const
{
	assert(context != EGL_NO_CONTEXT);
	eglSwapInterval(display, i);
}
