// stdlib includes
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <assert.h>
#include <GL/gl.h>
#include <boost/program_options.hpp>

namespace po = boost::program_options;

// include proper GL connector
#include "glwindow.h"
#if defined(USE_GLX)
#include "glxbackend.h"
GLBackend *createGLBackend()
{
	return new GLXBackend();
}
#elif defined(USE_EGL)
#include "eglbackend.h"
GLBackend *createGLBackend()
{
	return new EGLBackend();
}
#else
#error "No GL window type selected"
#endif

// configuration
const GLfloat boxWidth = 0.045f;
const GLfloat boxSpeed = 1.25f; // per second

// profiler
const int numProfilerStates = 5;
const char *profilerStateNames[numProfilerStates] = { "Pre-Render", "Drawing", "Swapping", "Post-Render", "Outside renderer"};

// utility functions
static double getTime()
{
	struct timespec tp;
	clock_gettime(CLOCK_MONOTONIC, &tp);
	return tp.tv_sec + 1e-9 * tp.tv_nsec;
}

static void drawRect(GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2)
{
	glVertex2f(x1, y1); glVertex2f(x2, y1); glVertex2f(x2, y2); glVertex2f(x1, y2);
}

// the window
class TearTestWindow : public GLWindow {
public:
	TearTestWindow() : GLWindow(XOpenDisplay(0), createGLBackend()), boxPos(0), boxDirection(1)
	{}

	void setSwapInterval(int i) {
		getBackend()->setSwapInterval(i);
	}

protected:
	virtual void initGL()
	{
		// initialize GL proper
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glDisable(GL_DEPTH_TEST);
		// initialize clocks
		lastFrame = getTime();
		// initailize profiler
		framect = 0;
		memset(stateTime, 0, sizeof(stateTime));
		curState = -1;
		lastDisplay = lastProfile = getTime();
	}
	
	virtual void resizeGL(unsigned int width, unsigned int height)
	{
		/* prevent divide-by-zero */                      
		if (height == 0)                                  
			height = 1;                                   
		glViewport(0, 0, width, height);
		glMatrixMode(GL_PROJECTION);                      
		glLoadIdentity();                                 
		glOrtho (0, 1, 1, 0, 0, 1);
		glMatrixMode(GL_MODELVIEW);      
		glClear(GL_COLOR_BUFFER_BIT);                                                 
		glFlush();
	}
	
	void profilerTick(int nextState)
	{
		assert (nextState >= 0 && nextState < numProfilerStates);
		double time = getTime();
		if (curState >= 0)
			stateTime[curState] += time-lastProfile;
		curState = nextState;
		lastProfile = time;
		// display?
		const double elapsed = time-lastDisplay;
		if (elapsed >= 3) {
			printf("%.1f fps, time spent: ", framect/elapsed);
			for (int i = 0; i < numProfilerStates; ++i) {
				if (i != 0) printf(", ");
				printf("%s %.1f%%", profilerStateNames[i], stateTime[i]/elapsed*100);
			}
			printf("\n");
			lastDisplay = time;
			framect = 0;
			memset(stateTime, 0, sizeof(stateTime));
		}
	}
	
	void renderGL()
	{                            
		profilerTick(0);
		double time = getTime();
		// anim
		double passedTime = time-lastFrame;
		boxPos += boxSpeed*passedTime*boxDirection;
		while (boxPos < 0 || boxPos+boxWidth > 1) { // wrapover
			if (boxPos < 0) {
				boxPos = -boxPos;
				boxDirection = -boxDirection;
			}
			else {
				boxPos = 1.0-boxWidth-(boxPos+boxWidth-1.0);
				boxDirection = -boxDirection;
			}
		}
		lastFrame = time;
		// draw
		//glFlush();
		profilerTick(1);
		//glClear(GL_COLOR_BUFFER_BIT);
		glBegin(GL_QUADS);
		// clear manually
		glColor3f(0.0f, 0.0f, 0.0f);
		drawRect(0, 0, 1, 1);
		glColor3f(0.8f, 1.0f, 0.75f);
		drawRect(boxPos, 0, boxPos+boxWidth, 1);
		glEnd();
		usleep(20*1000);
		profilerTick(2);
		getBackend()->swapBuffers();
// 		glDrawBuffer(GL_FRONT);
// 		int xpos = 0;
// 		int ypos = 0;
		//foreach (const QRect &r, region.rects()) {
			// convert to OpenGL coordinates
			//int y = displayHeight() - 0 - r.height();
// 			glBitmap(0, 0, 0, 0, 0 - xpos, 0 - ypos, NULL); // not glRasterPos2f, see glxbackend.cpp
// 			xpos = 0;
// 			ypos = 0;
// 			glCopyPixels(0, 0, getWidth(), getHeight(), GL_COLOR);
		//}
// 		glBitmap(0, 0, 0, 0, -xpos, -ypos, NULL); // move position back to 0,0
// 		glDrawBuffer(GL_BACK);
		profilerTick(3);
		glFlush();
		++framect;
		profilerTick(4);
	}
	
	virtual void handleKeyPress(KeySym key)
	{
		switch (key) {
			case XK_Escape: close(); break;
			case XK_F1: setFullscreen(!getFullscreen()); break;
			default: break;
		}
	}

private:
	double lastFrame;
	GLfloat boxPos, boxDirection;
	// FPS
	double lastDisplay, lastProfile;                 
	int framect, curState;
	double stateTime[numProfilerStates];
};



int main(int argc, char ** argv)
{
	// program options handling
	po::options_description desc("Allowed options");
	desc.add_options()
		("help,h", "produce help message")
		("swap-interval,i", po::value<int>(), "set swap interval")
	;
	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, desc), vm);
	po::notify(vm);

	if (vm.count("help")) {
		std::cout << desc << "\n";
		return 1;
	}

	// actual program
	TearTestWindow w;
	w.open(800, 600);
	if (vm.count("swap-interval"))
		w.setSwapInterval(vm["swap-interval"].as<int>());
	w.exec();
}
